from kevin_toolbox.computer_science.algorithm.utils import get_value_by_name


def set_value_by_name(var, name, value):
    """
        通过解释名字得到取值方式，然后到 var 中将对应部分的值修改为 value。

        参数：
            var:            任意支持索引取值的变量
            name:           <string> 取值方式
                                由多组 "<解释方式><键>" 组成。
                                解释方式支持以下几种:
                                    "@"     使用 eval() 读取键
                                    ":"     使用 str() 读取键
                                    "|"     依次尝试 str() 和 eval() 两种方式
                                示例:
                                    "@100"      表示指向 var[eval("100")]
                                    ":epoch"    表示指向 var["epoch"]
                                    "|1+1"    表示首先尝试指向 var["1+1"]，若不成功则尝试指向 var[eval("1+1")]
                                假设 var=dict(acc=[0.66,0.78,0.99])，如果你想将 var["acc"][1] 设置为 100，那么可以将 name 写成：
                                    ":acc@1" 或者 "|acc|1" 等。
                                注意，在 name 的开头也可以添加任意非解释方式的字符，本函数将直接忽略它们，比如下面的:
                                    "var:acc@1" 和 "xxxx|acc|1" 也能正常写入。
            value:          待赋给的值
    """
    assert isinstance(name, (str,))
    key = name
    for char in ":@|":
        key = key.rsplit(char, 1)[-1]
    assert len(key) < len(name), f'invalid name: {name}'

    item = get_value_by_name(var=var, name=name[:-1 - len(key)])

    flag = name[-1 - len(key)]
    if flag == "@":
        key = eval(key)
    elif flag == "|":
        try:
            _ = item[key]
        except:
            key = eval(key)

    item[key] = value
    return var


if __name__ == '__main__':
    print(set_value_by_name(var=dict(acc=[0.66, 0.78, 0.99]), name="|acc|0", value=1))
    print(set_value_by_name(var=dict(acc=[0.66, 0.78, 0.99]), name="www|acc|0", value=1))



def traverse(var, match_cond, action_mode="remove", converter=None):
    """
        遍历 var 找到符合 match_cond 的元素，将其按照 action_mode 指定的操作进行处理

        参数：
            var:                待处理数据
                                    当 var 不是 dict 或者 list 时，将直接返回 var 而不做处理
            match_cond:         <func> 元素的匹配条件
                                    函数类型为 def(parent_type, idx, value): xxxx
                                    其中：
                                        parent_type     该元素源自哪种结构体，有两个可能传入的值： list，dict
                                        idx             该元素在结构体中的位置，对于列表是 index，对于字典是 key
                                        value           元素的值
            action_mode:        <str> 如何对匹配上的元素进行处理
                                    目前支持：
                                        "remove"        将该元素移除
                                        "replace"       将该元素替换为 converter() 处理后的结果
            converter:          <func> 参见 action_mode 中的 "replace" 模式
                                    函数类型为 def(idx, value): xxxx
                                    其中 idx 和 value 的含义参见参数 match_cond 介绍
    """
    assert callable(match_cond)
    assert action_mode in {"replace", "remove"}
    if action_mode == "replace":
        assert callable(converter)

    return recursive_(var, match_cond, action_mode, converter)


def recursive_(var, match_cond, action_mode, converter):
    if isinstance(var, (list, dict)):
        items = reversed(list(enumerate(var))) if isinstance(var, list) else list(var.items())
        for k, v in items:
            if match_cond(type(var), k, v):
                if action_mode == "remove":
                    var.pop(k)
                else:
                    var[k] = converter(k, v)
            else:
                var[k] = recursive_(v, match_cond, action_mode, converter)
    else:
        pass
    return var


if __name__ == '__main__':
    import numpy as np

    x = [
        dict(d=3, c=4),
        np.array([[1, 2, 3]])
    ]

    y1 = traverse(var=x, match_cond=lambda _, k, v: type(v) is np.ndarray, action_mode="replace",
                  converter=lambda k, v: v.tolist())
    print(y1)

    y2 = traverse(var=y1, match_cond=lambda _, k, v: v == 3, action_mode="remove")
    print(y2)
